/******************************************************************************
 * $Header: /boot/home/agmsmith/Programming/AGMSRAMDiskDevice/RCS/AGMSRAMDiskDevice.c,v 1.19 2001/05/14 21:50:44 agmsmith Exp $
 *
 * This is a RAM Disk device for BeOS 5.0.  It simulates a real disk drive
 * using the computer's memory rather than real media.  Well, it doesn't
 * simulate the slow speed of a real drive.  It isn't necessarily really RAM
 * either, but can be virtual memory, which can end up being as slow as a real
 * disk.  But you can eject the media and otherwise pretend that it is like a
 * really big fast floppy disk or CD-ROM.
 *
 * Why am I writing this?  I'm just trying to get the icon to show up in the
 * desktop mount menu and changing the RAM disk to avoid requiring manual use
 * of DriveSetup to format the new RAM disk.  It turns out that Earl and the
 * guys have mkbfs in their system bootup script to do the job, and a similar
 * copy to zip file operation in the shutdown script.  I'll try to initialise
 * memory to a preformatted disk volume instead, using gzip compression.
 *
 * It is inspired by Earl Colby Pottinger's version which is based on the
 * one in the November 26, 1997 Be NewsLetter #101 by Dmitriy Budko
 * http://www-classic.be.com/aboutbe/benewsletter/Issue101.html#Insight2
 * with improvements along the way by "petersens" and Carlos Hasan.
 *
 * My (Alexander G. M. Smith) version of the RAM Disk device is released to
 * the public domain.  Use at your own risk, etc.  Send e-mail with suggestions
 * for improvement to agmsmith@achilles.net.
 *
 * $Log: AGMSRAMDiskDevice.c,v $
 * Revision 1.19  2001/05/14 21:50:44  agmsmith
 * Added compression level parameter so you can store the image file
 * in an uncompressed format if you wish (suggested by Olaf Roß).
 * Also added an fsync before closing the file, to make sure that
 * it gets flushed out to the real disk during shutdown.
 *
 * Revision 1.18  2001/01/03 03:14:52  agmsmith
 * Added hack for detecting system shutdown and writing out
 * the disk image at the last moment.
 *
 * Revision 1.17  2001/01/02 03:06:06  agmsmith
 * Temporary version that dumps to disk image when number of
 * open files drops to zero, doesn't work since BeOS keeps one
 * file open even when it is rebooting.
 *
 * Revision 1.16  2001/01/02 00:13:51  agmsmith
 * *** empty log message ***
 *
 * Revision 1.15  2001/01/02 00:11:15  agmsmith
 * Only load and decompress the disk image when the first open is
 * done, to avoid redundant time wasting decompressions during
 * boot-up (init gets called several times).
 *
 * Revision 1.14  2001/01/01 23:04:05  agmsmith
 * Now loads and saves the disk contents from/to a disk image
 * file when the disk is "ejected".  Also uses ZLib's deflate
 * code to compress the disk image file.
 *
 * Revision 1.13  2000/12/30 19:10:48  agmsmith
 * Added read-only and write-once disk emulation.
 *
 * Revision 1.12  2000/12/28 00:33:34  agmsmith
 * Now with resizeable media, you can reallocate the memory using
 * a certain IOCtl, while the "media" is inserted.  Also had to
 * add a drive not ready feature - for when the memory area isn't
 * allocated (due to asking for too much memory and having it fail).
 *
 * Revision 1.11  2000/12/27 22:28:30  agmsmith
 * Code reorganising in preparation for resizing feature, also
 * added non-removeable option.
 *
 * Revision 1.10  2000/12/27 19:56:04  agmsmith
 * Added ability to remove debug messages.  Also now detects closed
 * files and prevents furthur operations on them.
 *
 * Revision 1.9  2000/12/15 22:51:47  agmsmith
 * Now reads its parameters from the device configuration file.
 *
 * Revision 1.8  2000/12/14 19:25:03  agmsmith
 * Naming changes and clearer output messages.
 *
 * Revision 1.7  2000/11/11 17:54:57  agmsmith
 * Under construction...
 *
 * Revision 1.6  2000/10/31 03:10:11  agmsmith
 * Still making it removable.  It sort of works, a bit.
 *
 * Revision 1.5  2000/10/31 01:07:46  agmsmith
 * Adding removeable media to RAM disk simulation.
 *
 * Revision 1.4  2000/10/30 18:33:03  agmsmith
 * More typing.
 *
 * Revision 1.3  2000/10/23 19:53:45  baron
 * Initial comments.
 *
 * Revision 1.2  2000/10/23 19:28:28  baron
 * Adding RCS headers.
 */

#include <KernelExport.h>
#include <Drivers.h>
#include <Errors.h>
#include <driver_settings.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

/* We use the ZLib 1.1.3 compression library, which lets you provide malloc,
so it can work in the kernel address space.  Fortunately BeOS 5.00 comes with
a precompiled version of the library, so all we have to do is to include it
and add libz.a to the project settings. */

#include <zlib.h>


#define DPRINT_NON_ERROR_MESSAGES 0
#define DPRINT_READ_AND_WRITE_OPERATIONS 0
  /* 0 or 1.  1 turns on the printing (via dprintf - to the system log)
  of debug messages for normal operations, which can make things slow.
  0 makes the code only print error conditions and special messages. */


int32 api_version = B_CUR_DRIVER_API_VERSION;
  /* Exported value so that the OS knows which driver API version we
  compiled with. */


/* List of devices in the /dev hierarchy which we support. */

const char *PublishedDeviceNames [] =
{
  "disk/virtual/AGMSRAMDiskDevice/raw", /* Need "/raw" for disk partitions. */
  NULL
};


/* Values set from the configuration file by ReadConfigurationParameters(). */

static long InitialMemoryByteSize;
static bool InitialUseVirtualMemory;
static bool DiskIsRemoveable; /* true for eject allowed. */
static bool DiskIsReadOnly; /* true to disable write ability. */
static bool DiskIsWriteOnce; /* true to simulate a write once disk. */
static bool InsertDiskAtStartup; /* false to start with media ejected. */
static bool UseSafeTemporaryFileWhenSavingCompressedDiskImage;
static bool NeverWriteCompressedDiskImage; /* for repeatable RAM disk init. */
static int  CompressionLevel; /* 0 for none, 9 for maximum compression. */
static int  FlushAtNumberOfOpenFilesHack; /* for detecting BeOS shutdown. */
static char PathNameForCompressedDiskImage [B_PATH_NAME_LENGTH];
static char PathNameForTempDiskImage [B_PATH_NAME_LENGTH];


/* Globals identifying the memory area we use.  It's allocated at
   initialisation time and remains allocated forever.  Of course,
   these variables need to get reinitialised when the driver is
   reloaded, based on the eternal memory area's properties. */

static const char *MemoryAreaName = "AGMSRAMDiskDevice Contents";
static void *MemoryAreaPntr = NULL;
  /* RAM disk contents stored here, can be NULL, which shows up as
  device not ready. */
static unsigned int MemorySize; /* Number of bytes in RAM disk area. */
static bool MemoryIsDirty; /* true if has been written to since loaded. */
#define FAKE_BLOCK_SIZE 512 /* 512 is needed for Intel partition tables. */

/* Other globals.  These get allocated when the driver is (re)loaded
   and deallocated when the driver is unloaded. */

static const char *MainSemaphoreName = "AGMSRAMDiskDevice Main Semaphore";
static sem_id MainSemaphoreID = -1; /* Guards all critical sections. */
static volatile unsigned int CookieCounter; /* Used in cookies from "open". */
static volatile bool DiskIsInserted = false; /* true when disk is "loaded". */
static volatile unsigned int MediaLoadedSerialNo; /* +1 at each load/unload. */
static volatile unsigned int NumberOfOpenFiles; /* Mostly for shutdown hack. */
static volatile unsigned int RealFlushCountdown; /* Hack avoids extra flush. */

#define FILEBUFFERSIZE (1024 * 1024)
  /* Size of buffer used for file I/O during compression / decompression. */


/* Each open file instance gets one of these allocated (using kernel malloc
   which means it gets locked into kernel memory) and the cookie given to the
   user points to this structure. */

typedef struct CookieStruct
{
  uint32 magicValue; /* Set to 'AGMS' to show this is a valid structure. */
  void *selfPntr; /* Address of start of this structure - a validity check. */
  unsigned int cookieCounter; /* So we can tell one open from another. */
  unsigned int mediaSerial; /* Value of MediaLoadedSerialNo at last use. */
  bool isOpen; /* Set to false in close() to force later operations to fail. */
  bool readAllowed; /* Reflects read / write mode when device opened. */
  bool writeAllowed;
} CookieRecord, *CookiePointer;


/* A few pictures of a RAM disk drive for the desktop to display. */

static const uchar DiskIcon32 [] = { /* The 32x32 pixel image. */
0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,

0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,

0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,

0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,

0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,

0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,

0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,

0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
0x03, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,

0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x03,
0x03, 0x05, 0x05, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,

0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x03, 0x05, 0x0E, 0x0A,
0x09, 0x0A, 0x11, 0x00, 0x05, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,

0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x03, 0x0E, 0x03, 0xD6, 0x04, 0x1A,
0x1A, 0xD6, 0x07, 0x0A, 0x05, 0x05, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,

0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x03, 0x05, 0x05, 0x1A, 0x1A, 0x1A, 0x07,
0x1A, 0x09, 0xAF, 0x07, 0x07, 0x11, 0x00, 0x05, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,

0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x04, 0x0E, 0x03, 0x1A, 0x1A, 0x0C, 0x1A, 0x0A, 0x1A,
0x08, 0x09, 0x07, 0x0A, 0x0C, 0x06, 0x08, 0x11, 0x11, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,

0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x04, 0x05, 0x09, 0x08, 0x09, 0x0C, 0x0B, 0x0E, 0x1A, 0x1A,
0xAF, 0x08, 0x0A, 0x1A, 0x1A, 0x0D, 0x09, 0xD6, 0x06, 0x00, 0x05, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,

0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x0E, 0x04, 0xD6, 0x0C, 0x09, 0x0B, 0x08, 0x09, 0x0A, 0x1A, 0xAF,
0x0D, 0x0F, 0x1A, 0x0A, 0x09, 0x1A, 0x0D, 0x0D, 0x08, 0x04, 0x04, 0x11, 0xFF, 0xFF, 0xFF, 0xFF,

0xFF, 0xFF, 0xFF, 0xFF, 0x05, 0x06, 0x0B, 0x0C, 0x0B, 0x0B, 0x0D, 0x0C, 0x08, 0x1A, 0x07, 0x0A,
0x1A, 0x0B, 0x1A, 0x1A, 0x0A, 0x1A, 0x0A, 0x0A, 0x0A, 0x0A, 0x08, 0x09, 0x00, 0xFF, 0xFF, 0xFF,

0xFF, 0xFF, 0xFF, 0xFF, 0x07, 0x0A, 0x08, 0x0C, 0x06, 0x06, 0x0C, 0x0B, 0x0C, 0x0C, 0x07, 0x1A,
0x19, 0x09, 0x08, 0x1A, 0x0B, 0x0A, 0x0A, 0x0C, 0x0D, 0x0B, 0x0A, 0x04, 0x04, 0x00, 0xFF, 0xFF,

0xFF, 0xFF, 0xFF, 0xFF, 0x09, 0x0B, 0x09, 0x04, 0x0C, 0x0C, 0x0C, 0x0B, 0x0A, 0x0B, 0xAF, 0x0C,
0x1A, 0x0C, 0x1A, 0x1A, 0x0A, 0x0A, 0x0A, 0x0B, 0x0A, 0x06, 0x06, 0x09, 0x07, 0x03, 0xFF, 0xFF,

0xFF, 0xFF, 0xFF, 0xFF, 0x12, 0x61, 0x08, 0x09, 0xD6, 0x08, 0x07, 0xD6, 0x0C, 0x0C, 0x09, 0x0A,
0x0C, 0x1A, 0x1A, 0x0A, 0x0D, 0x0C, 0x0B, 0x07, 0x06, 0x0C, 0x0B, 0x14, 0x14, 0x03, 0xFF, 0xFF,

0xFF, 0xFF, 0xFF, 0x12, 0x61, 0x0B, 0x0A, 0x09, 0x09, 0x07, 0xD6, 0x0A, 0x0C, 0x0B, 0x09, 0x0A,
0x0A, 0x09, 0x08, 0x09, 0x0C, 0x0A, 0x08, 0x08, 0x0A, 0x0B, 0x1A, 0x1A, 0x14, 0x03, 0xFF, 0xFF,

0xFF, 0xFF, 0x12, 0x12, 0x12, 0x12, 0x0C, 0x12, 0x61, 0x0B, 0x0B, 0x0A, 0xD6, 0x06, 0x0B, 0x0A,
0x09, 0x08, 0xAF, 0x0A, 0x0A, 0x07, 0x0B, 0x0A, 0x0A, 0x07, 0x07, 0x12, 0x12, 0x10, 0xFF, 0xFF,

0xFF, 0xFF, 0x12, 0x11, 0xD6, 0xFF, 0x0A, 0x61, 0x61, 0x09, 0x09, 0x08, 0x09, 0x09, 0x08, 0x0A,
0x09, 0x0B, 0x0B, 0x07, 0x06, 0x09, 0x0B, 0x1A, 0x12, 0x14, 0x07, 0x04, 0x12, 0x10, 0xFF, 0xFF,

0xFF, 0xFF, 0x16, 0x12, 0xFF, 0xFF, 0x06, 0x12, 0x0C, 0x0A, 0x12, 0x61, 0x0A, 0x09, 0x09, 0x05,
0x09, 0x0B, 0xD6, 0x04, 0x0B, 0x0C, 0x0A, 0x1A, 0x12, 0x12, 0x10, 0xFF, 0x12, 0x10, 0x0A, 0xFF,

0xFF, 0x16, 0x16, 0x16, 0xFF, 0xFF, 0x12, 0x11, 0x08, 0x0C, 0x12, 0x61, 0x0B, 0x09, 0x0A, 0x0D,
0x0D, 0x07, 0x0A, 0x0B, 0x09, 0x1A, 0x12, 0x14, 0x04, 0x12, 0x10, 0xFF, 0x12, 0x0E, 0x0E, 0xFF,

0xFF, 0x18, 0x16, 0x1E, 0xFF, 0x12, 0x12, 0x11, 0xFF, 0xD6, 0x12, 0x11, 0x0B, 0x0A, 0x11, 0x61,
0x0B, 0x04, 0x09, 0x1A, 0x14, 0x09, 0x1A, 0x10, 0x0A, 0x12, 0x12, 0xFF, 0xFF, 0x0D, 0x0D, 0xFF,

0xFF, 0xFF, 0x18, 0xFF, 0xFF, 0x16, 0x12, 0xFF, 0xFF, 0x12, 0x10, 0x11, 0x07, 0x0B, 0x12, 0x61,
0x0B, 0x09, 0x0D, 0x1A, 0x14, 0x10, 0x12, 0x12, 0x10, 0x0A, 0x12, 0x10, 0x0D, 0x1E, 0x0D, 0xFF,

0xFF, 0xFF, 0xFF, 0xFF, 0x16, 0x16, 0x16, 0xFF, 0xFF, 0x12, 0x10, 0xFF, 0x0A, 0x12, 0x11, 0x07,
0x0A, 0x07, 0x0D, 0x04, 0x12, 0x10, 0xFF, 0x12, 0x10, 0xFF, 0x18, 0x0D, 0x0E, 0x0D, 0xFF, 0xFF,

0xFF, 0xFF, 0xFF, 0xFF, 0x18, 0x16, 0x1E, 0xFF, 0xFF, 0x12, 0x11, 0xFF, 0xFF, 0x12, 0x11, 0x08,
0x0A, 0x09, 0x08, 0x0C, 0x12, 0x10, 0xFF, 0x12, 0x0D, 0xFF, 0xFF, 0x1E, 0x0D, 0xFF, 0xFF, 0xFF,

0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x18, 0xFF, 0xFF, 0xFF, 0x16, 0x11, 0xFF, 0xFF, 0x12, 0x11, 0xFF,
0x04, 0x08, 0x0A, 0xFF, 0x12, 0x10, 0xFF, 0x18, 0x0D, 0x0D, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,

0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x16, 0x16, 0x16, 0xFF, 0xFF, 0x16, 0x11, 0xFF,
0xFF, 0x08, 0xFF, 0xFF, 0x12, 0x0D, 0xFF, 0xFF, 0x1E, 0x0D, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,

0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x18, 0x16, 0x1E, 0xFF, 0x16, 0x16, 0x16, 0xFF,
0xFF, 0xFF, 0xFF, 0xFF, 0x18, 0x0E, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,

0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x18, 0xFF, 0xFF, 0xFF, 0x16, 0x1E, 0x18,
0xFF, 0xFF, 0xFF, 0xFF, 0x0E, 0x1E, 0x0E, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
};


static const uchar DiskIcon16 [] = { /* The smaller mini-icon 16x16 pixels. */
0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x05, 0x02, 0x02, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x04, 0x03, 0x07, 0x07, 0x06, 0x01, 0x03, 0xFF, 0xFF, 0xFF, 0xFF,
0xFF, 0xFF, 0xFF, 0xFF, 0x03, 0xD6, 0x0F, 0x10, 0x11, 0x09, 0x09, 0x01, 0x03, 0xFF, 0xFF, 0xFF,
0xFF, 0xFF, 0xFF, 0x04, 0x08, 0x12, 0x10, 0x16, 0x09, 0x0D, 0x0E, 0x0A, 0x06, 0x04, 0xFF, 0xFF,
0xFF, 0xFF, 0x07, 0x08, 0x0B, 0x0B, 0x0D, 0x0E, 0x10, 0x16, 0x12, 0x0B, 0x08, 0x0A, 0x01, 0xFF,
0xFF, 0xFF, 0x0A, 0x08, 0x09, 0x0B, 0x0B, 0x0E, 0x12, 0x15, 0x0A, 0x0B, 0x0A, 0x07, 0x03, 0xFF,
0xFF, 0x09, 0x14, 0x09, 0x08, 0x08, 0x0C, 0x0A, 0x0E, 0x0D, 0x0C, 0x08, 0x0A, 0x15, 0x0B, 0xFF,
0xFF, 0x12, 0x0B, 0x10, 0x11, 0x09, 0x08, 0x0A, 0x09, 0x0A, 0x08, 0x0E, 0x0E, 0x09, 0x11, 0xFF,
0xFF, 0x15, 0x05, 0x0E, 0x0A, 0x16, 0x0A, 0x09, 0x0A, 0x08, 0x0E, 0x13, 0x0F, 0xAF, 0x10, 0xFF,
0xFF, 0x15, 0x0E, 0x0E, 0x08, 0x11, 0x0A, 0x16, 0x09, 0x13, 0x10, 0x14, 0x0E, 0x11, 0x14, 0xFF,
0xFF, 0xFF, 0x17, 0x11, 0x0C, 0x0A, 0x0C, 0x0C, 0x09, 0x09, 0x11, 0x0D, 0x0D, 0x14, 0x0C, 0xFF,
0xFF, 0xFF, 0xFF, 0xFF, 0x13, 0x0E, 0x0D, 0x0E, 0xFF, 0xFF, 0x10, 0x0F, 0x11, 0x0A, 0xFF, 0xFF,
0xFF, 0xFF, 0xFF, 0xFF, 0x14, 0x0F, 0x14, 0x16, 0xFF, 0xFF, 0x15, 0x0C, 0x09, 0xFF, 0xFF, 0xFF,
} ;



/******************************************************************************
 * Read the startup parameters from the settings file, usually it will be
 * in the "/boot/home/config/settings/kernel/drivers/AGMSRAMDiskDevice" file.
 * The results are sanity checked and stored in global variables.
 */

static void ReadConfigurationParameters (void)
{
  const char *DefaultCompressionLevelString = "9";
  const char *DefaultFlushString = "1";
  const char *DefaultSizeString = "64000000";
  const char *EmptyString = "";
  const char *InitialSizeString = "initial_size_in_bytes";
  const char *InsertDiskAtStartupString = "insert_disk_at_startup";
  const char *IsRemoveableString = "disk_is_removeable";
  const char *IsReadOnlyString = "disk_is_read_only";
  const char *IsWriteOnceString = "disk_is_write_once";
  const char *NeverWriteString = "never_write_compressed_disk_image";
  const char *CompressionLevelString = "compression_level";
  const char *FlushNumberString = "number_of_open_files_for_shutdown_detection";
  const char *SafeTempString = "use_safe_temporary_file_when_saving_compressed_disk_image_file";
  const char *PathNameString = "path_name_of_compressed_disk_image_file";
  const char *VirtualMemoryString = "use_virtual_memory";
  const char *ParameterString;
  void       *SettingsPntr;

  /* Set up default values for the parameters, in case there isn't a
  settings file for them. */

  InitialMemoryByteSize = atol (DefaultSizeString);
  InitialUseVirtualMemory = true;
  DiskIsRemoveable = false;
  DiskIsReadOnly = false;
  DiskIsWriteOnce = false;
  InsertDiskAtStartup = true;
  UseSafeTemporaryFileWhenSavingCompressedDiskImage = false;
  NeverWriteCompressedDiskImage = false;
  CompressionLevel = 9;
  FlushAtNumberOfOpenFilesHack = 1;
  PathNameForCompressedDiskImage [0] = 0;

  SettingsPntr = load_driver_settings ("AGMSRAMDiskDevice");
  if (SettingsPntr != NULL)
  {
    ParameterString = get_driver_parameter (SettingsPntr,
      InitialSizeString, DefaultSizeString, DefaultSizeString);
    InitialMemoryByteSize = atol (ParameterString);

    InitialUseVirtualMemory = get_driver_boolean_parameter (SettingsPntr,
      VirtualMemoryString, true, true);

    DiskIsRemoveable = get_driver_boolean_parameter (SettingsPntr,
      IsRemoveableString, false, false);

    DiskIsReadOnly = get_driver_boolean_parameter (SettingsPntr,
      IsReadOnlyString, false, false);

    DiskIsWriteOnce = get_driver_boolean_parameter (SettingsPntr,
      IsWriteOnceString, false, false);

    InsertDiskAtStartup = get_driver_boolean_parameter (SettingsPntr,
      InsertDiskAtStartupString, true, true);

    UseSafeTemporaryFileWhenSavingCompressedDiskImage =
      get_driver_boolean_parameter (SettingsPntr,
      SafeTempString, false, false);

    NeverWriteCompressedDiskImage = get_driver_boolean_parameter (
      SettingsPntr, NeverWriteString, false, false);

    ParameterString = get_driver_parameter (SettingsPntr,
      CompressionLevelString, DefaultCompressionLevelString,
      DefaultCompressionLevelString);
    CompressionLevel = atoi (ParameterString);
    if (CompressionLevel < Z_NO_COMPRESSION)
      CompressionLevel = Z_NO_COMPRESSION;
    if (CompressionLevel > Z_BEST_COMPRESSION)
      CompressionLevel = Z_BEST_COMPRESSION;

    ParameterString = get_driver_parameter (SettingsPntr,
      FlushNumberString, DefaultFlushString, DefaultFlushString);
    FlushAtNumberOfOpenFilesHack = atoi (ParameterString);

    ParameterString = get_driver_parameter (SettingsPntr,
      PathNameString, EmptyString, EmptyString);
    strncpy (PathNameForCompressedDiskImage, ParameterString,
      sizeof (PathNameForCompressedDiskImage));
    PathNameForCompressedDiskImage[sizeof (PathNameForCompressedDiskImage) - 5]
      = 0; /* Truncate the string, leaving room to append ".tmp". */

    unload_driver_settings (SettingsPntr);
  }
  else
  {
    dprintf ("AGMSRAMDiskDevice: Unable to read settings file, usually it is "
      "\"/boot/home/config/settings/kernel/drivers/AGMSRAMDiskDevice\", "
      "defaults will be used instead.\n");
  }

  if (InitialMemoryByteSize < FAKE_BLOCK_SIZE)
    InitialMemoryByteSize = atol (DefaultSizeString);

  /* Round up to a multiple of FAKE_BLOCK_SIZE.  The actual allocation later
  will get rounded up to the next memory page size too. */

  InitialMemoryByteSize = (InitialMemoryByteSize + FAKE_BLOCK_SIZE - 1) /
    FAKE_BLOCK_SIZE * FAKE_BLOCK_SIZE;

  /* Generate the other file names from the disk image one. */

  strcpy (PathNameForTempDiskImage, PathNameForCompressedDiskImage);
  strcat (PathNameForTempDiskImage, ".tmp");

  /* Echo the settings back to the log file, with enough info to recreate the
  settings file if none exists (the actual names of the paramters). */

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("AGMSRAMDiskDevice: Settings: %s %ld\n",
    InitialSizeString, InitialMemoryByteSize);
  dprintf ("AGMSRAMDiskDevice: Settings: %s %s\n",
    VirtualMemoryString, InitialUseVirtualMemory ? "yes" : "no");
  dprintf ("AGMSRAMDiskDevice: Settings: %s %s\n",
    IsRemoveableString, DiskIsRemoveable ? "yes" : "no");
  dprintf ("AGMSRAMDiskDevice: Settings: %s %s\n",
    IsReadOnlyString, DiskIsReadOnly ? "yes" : "no");
  dprintf ("AGMSRAMDiskDevice: Settings: %s %s\n",
    IsWriteOnceString, DiskIsWriteOnce ? "yes" : "no");
  dprintf ("AGMSRAMDiskDevice: Settings: %s %s\n",
    InsertDiskAtStartupString, InsertDiskAtStartup ? "yes" : "no");
  dprintf ("AGMSRAMDiskDevice: Settings: %s %s\n",
    SafeTempString,
    UseSafeTemporaryFileWhenSavingCompressedDiskImage ? "yes" : "no");
  dprintf ("AGMSRAMDiskDevice: Settings: %s %s\n",
    NeverWriteString, NeverWriteCompressedDiskImage ? "yes" : "no");
  dprintf ("AGMSRAMDiskDevice: Settings: %s %d\n",
    CompressionLevelString, CompressionLevel);
  dprintf ("AGMSRAMDiskDevice: Settings: %s %d\n",
    FlushNumberString, FlushAtNumberOfOpenFilesHack);
  dprintf ("AGMSRAMDiskDevice: Settings: %s \"%s\"\n",
    PathNameString, PathNameForCompressedDiskImage);
#endif
}



/******************************************************************************
 * Allocates the memory used for the RAM disk.  If it is already present, it
 * just updates the globals (MemoryAreaPntr, MemorySize) to reflect the
 * existing memory size.  Uses InitialMemoryByteSize and
 * InitialUseVirtualMemory when allocating the memory.  Returns an error code
 * if it fails.
 */

static status_t AllocateDiskContentsArea (void)
{
  area_id   AreaID;
  area_info AreaInfo;
  status_t  ErrorCode;

  MemorySize = (InitialMemoryByteSize + B_PAGE_SIZE - 1) /
    B_PAGE_SIZE * B_PAGE_SIZE; /* Round up to next multiple of page size. */
  MemoryAreaPntr = NULL;

  AreaID = find_area (MemoryAreaName);

  /* Allocate the new area if we didn't find the old one. */

  if (AreaID < 0)
  {
    AreaID = create_area (MemoryAreaName, &MemoryAreaPntr,
      B_ANY_KERNEL_ADDRESS, MemorySize,
      InitialUseVirtualMemory ? B_NO_LOCK : B_FULL_LOCK,
      B_READ_AREA | B_WRITE_AREA);
    if (AreaID < 0)
    {
      dprintf ("AGMSRAMDiskDevice: AllocateDiskContentsArea failed to "
        "allocate %d bytes of memory, error code $%X.\n", MemorySize,
        (unsigned int) AreaID);
      MemoryAreaPntr = NULL;
      return B_DEV_NO_MEMORY;
    }

#if DPRINT_NON_ERROR_MESSAGES
    dprintf ("AGMSRAMDiskDevice: AllocateDiskContentsArea allocated %u bytes "
      "in memory area \"%s\" with area ID of %d.  Virtual memory (swapping) "
      "is %s.\n", MemorySize, MemoryAreaName, (int) AreaID,
      InitialUseVirtualMemory ? "on" : "off");
#endif
  }

  /* Get information about the area (new or old) and set
  the globals describing the area. */

  memset (&AreaInfo, 0, sizeof (AreaInfo));
  ErrorCode = get_area_info (AreaID, &AreaInfo);
  if (ErrorCode < 0)
  {
    dprintf ("AGMSRAMDiskDevice: AllocateDiskContentsArea can't get "
      "area info, error $%lX.\n", ErrorCode);
    return ErrorCode;
  }

  MemoryAreaPntr = AreaInfo.address;
  MemorySize = AreaInfo.size / FAKE_BLOCK_SIZE * FAKE_BLOCK_SIZE;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("AGMSRAMDiskDevice: AllocateDiskContentsArea found memory "
    "area ID %ld, address $%lX, size %ld, name \"%s\", %ld bytes currently "
    "swapped in RAM.\n", (long) AreaID, (unsigned long) MemoryAreaPntr,
    (long) MemorySize, AreaInfo.name, (long) AreaInfo.ram_size);
#endif

  return B_OK;
}



/******************************************************************************
 * Deallocates the RAM disk memory area, if it exists.
 */

static void FreeDiskContentsArea (void)
{
  area_id AreaID;

  AreaID = find_area (MemoryAreaName);
  if (AreaID >= 0)
  {
#if DPRINT_NON_ERROR_MESSAGES
    dprintf ("AGMSRAMDiskDevice: FreeDiskContentsArea found existing RAM disk "
      "memory area ID %ld, deleting it.\n", AreaID);
#endif
    MemoryAreaPntr = NULL;
    delete_area (AreaID);
  }
}



/******************************************************************************
 * Check for write once disks if the given range of storage space is
 * writeable, returns false if it already contains data.  Assumes that
 * the input range is valid, otherwise it will crash.
 */

static bool CheckForEmptyWriteOnceData (off_t Position, size_t Amount)
{
  unsigned char *BytePntr;
  int            i;

  BytePntr = MemoryAreaPntr + Position;
  for (i = Amount; i > 0; i--)
  {
    if (*BytePntr++ != 0)
      break;
  }

  return (i <= 0); /* TRUE if loop finished completely, all zeroes. */
}



/******************************************************************************
 * Utility function for the ZLib library to allocate memory.  We will just
 * use the kernel malloc, even though it does lock pages in memory etc.
 */

voidpf MyAllocForZLib (voidpf opaque, uInt items, uInt size)
{
  unsigned long Amount;
  void         *Pntr;

  Amount = items * (unsigned long) size;
  Pntr = malloc (Amount);
  if (Pntr == NULL)
  {
    dprintf ("AGMSRAMDiskDevice: MyAllocForZLib failed to "
      "allocate %lu bytes (%d items of %d size).\n",
      Amount, (int) items, (int) size);
    return Z_NULL;
  }

  return Pntr;
}



/******************************************************************************
 * Utility function for the ZLib library to deallocate memory.
 */

void MyFreeForZLib (voidpf opaque, voidpf address)
{
  free (address);
}



/******************************************************************************
 * Decompress the compressed disk image file into memory.  Think of it as
 * simulating inserting a virtual disk in the disk drive.  If the file doesn't
 * exist, or the memory hasn't been allocated then it does nothing.  Returns
 * true if successful.
 */

bool CopyDiskImageToMemory (void)
{
  int      AmountRead;
  int      ErrorCode;
  char    *FileBuffer = NULL;
  int      FileHandle = -1;
  bool     InflateInitWasDone = false;
  bool     ReturnCode = false;
  z_stream ZLibStream;

  if (MemoryAreaPntr == NULL)
  {
    dprintf ("AGMSRAMDiskDevice: CopyDiskImageToMemory doing nothing since "
      "the disk memory area hasn't been allocated yet.\n");
    goto NormalExit;
  }

  if (PathNameForCompressedDiskImage[0] == 0)
  {
    dprintf ("AGMSRAMDiskDevice: CopyDiskImageToMemory doing nothing since "
      "there isn't a name for the compressed disk image file.\n");
    goto NormalExit;
  }

  FileHandle = open (PathNameForCompressedDiskImage,
    O_RDONLY | O_BINARY, 0666);
  if (FileHandle < 0)
  {
    ErrorCode = errno;
    dprintf ("AGMSRAMDiskDevice: CopyDiskImageToMemory unable to open file "
      "\"%s\", error code %d (%s).\n",
      PathNameForCompressedDiskImage, ErrorCode, strerror (ErrorCode));
    goto NormalExit;
  }

  FileBuffer = malloc (FILEBUFFERSIZE);
  if (FileBuffer == NULL)
  {
    dprintf ("AGMSRAMDiskDevice: CopyDiskImageToMemory failed to allocate "
      "memory for the file reading buffer.\n");
    goto NormalExit;
  }

  memset (&ZLibStream, 0, sizeof (ZLibStream));
  ZLibStream.zalloc = MyAllocForZLib;
  ZLibStream.zfree = MyFreeForZLib;

  /* Read some initial data to fill the buffer, so that ZLib can read the
  file header to determine the compression method and thus allocate memory
  for its own use. */

  AmountRead = read (FileHandle, FileBuffer, FILEBUFFERSIZE);
  if (AmountRead < 0)
  {
    ErrorCode = errno;
    dprintf ("AGMSRAMDiskDevice: CopyDiskImageToMemory unable to read "
      "\"%s\", error code %d (%s).\n",
      PathNameForCompressedDiskImage, ErrorCode, strerror (ErrorCode));
    goto NormalExit;
  }
  if (AmountRead == 0)
  {
    dprintf ("AGMSRAMDiskDevice: CopyDiskImageToMemory "
      "\"%s\" is an empty file, can't do anything with it.\n",
      PathNameForCompressedDiskImage);
    goto NormalExit;
  }

  ZLibStream.next_in = FileBuffer;
  ZLibStream.avail_in = AmountRead;

  /* And of course the output fills the memory area. */

  ZLibStream.next_out = MemoryAreaPntr;
  ZLibStream.avail_out = MemorySize;

  ErrorCode = inflateInit (&ZLibStream);
  if (ErrorCode != Z_OK)
  {
    dprintf ("AGMSRAMDiskDevice: CopyDiskImageToMemory "
      "inflateInit for \"%s\" failed, code %d (%s).\n",
      PathNameForCompressedDiskImage, ErrorCode,
      ZLibStream.msg ? ZLibStream.msg : "?");
    goto NormalExit;
  }
  InflateInitWasDone = true;

  /* Decompress data until there is no more input or until output is full. */

  while (true)
  {
    if (ZLibStream.avail_out == 0)
    {
      dprintf ("AGMSRAMDiskDevice: CopyDiskImageToMemory ran out of space "
        "for output, perhaps the input disk image \"%s\" is for a bigger "
        "disk?\n", PathNameForCompressedDiskImage);
      goto NormalExit;
    }

    if (ZLibStream.avail_in == 0)
    {
      /* Read some more data if possible. */

      AmountRead = read (FileHandle, FileBuffer, FILEBUFFERSIZE);
      if (AmountRead < 0)
      {
        ErrorCode = errno;
        dprintf ("AGMSRAMDiskDevice: CopyDiskImageToMemory unable to read "
          "\"%s\", error code %d (%s).\n",
          PathNameForCompressedDiskImage, ErrorCode, strerror (ErrorCode));
        goto NormalExit;
      }

      ZLibStream.next_in = FileBuffer;
      ZLibStream.avail_in = AmountRead;
    }

    /* Do a portion of the decompression. */

    ErrorCode = inflate (&ZLibStream, Z_NO_FLUSH);

    if (ErrorCode == Z_STREAM_END)
    {
      if (ZLibStream.avail_out > 0)
      {
        dprintf ("AGMSRAMDiskDevice: CopyDiskImageToMemory finished early, "
          "there were %d bytes of memory not initialised by the data from "
          "\"%s\", perhaps the file was for a smaller disk size?\n",
          ZLibStream.avail_out, PathNameForCompressedDiskImage);
        goto NormalExit;
      }
      break; /* Finished decompressing all data. */
    }

    if (ErrorCode != Z_OK)
    {
      dprintf ("AGMSRAMDiskDevice: CopyDiskImageToMemory "
        "inflate for \"%s\" failed after reading %ld bytes, code %d (%s).\n",
        PathNameForCompressedDiskImage, ZLibStream.total_in,
         ErrorCode, ZLibStream.msg ? ZLibStream.msg : "?");
      goto NormalExit;
    }
  }

  /* Successful. */

  ReturnCode = true;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("AGMSRAMDiskDevice: CopyDiskImageToMemory "
    "successfully decompressed %ld bytes from \"%s\" into %ld bytes "
    "(%d%%).\n",
    ZLibStream.total_in, PathNameForCompressedDiskImage, ZLibStream.total_out,
    (ZLibStream.total_in > 0)
    ? (int) (ZLibStream.total_out * 100LL /* LL for 64 bit ints */ /
      ZLibStream.total_in)
    : 0);
#endif

  /* Clean up outstanding allocations. */

NormalExit:
  if (InflateInitWasDone)
  {
    ErrorCode = inflateEnd (&ZLibStream);
    if (ErrorCode != Z_OK)
    {
      dprintf ("AGMSRAMDiskDevice: CopyDiskImageToMemory inflateEnd for "
      "\"%s\" returned error code %d (%s), possibly data is corrupt or "
      "image size is bigger than the disk size.\n",
      PathNameForCompressedDiskImage, ErrorCode,
      ZLibStream.msg ? ZLibStream.msg : "?");
    }
  }
  if (FileHandle >= 0)
    close (FileHandle);
  if (FileBuffer != NULL)
    free (FileBuffer);
  return ReturnCode;
}



/******************************************************************************
 * Copy the data in the RAM disk into an image file, using ZLib compression.
 * If there are problems creating the file, or the memory hasn't been
 * allocated then it does nothing.  Returns true if successful.
 */

bool CopyMemoryToDiskImage (void)
{
  int      AmountWritten;
  int      ErrorCode;
  char    *FileBuffer = NULL;
  int      FileHandle = -1;
  bool     DeflateInitWasDone = false;
  char    *OpenFileName = NULL;
  bool     ReturnCode = false;
  z_stream ZLibStream;

  if (NeverWriteCompressedDiskImage)
  {
    dprintf ("AGMSRAMDiskDevice: CopyMemoryToDiskImage doing nothing since "
      "the user doesn't want to write to the disk image file.\n");
    goto NormalExit;
  }

  if (MemoryAreaPntr == NULL)
  {
    dprintf ("AGMSRAMDiskDevice: CopyMemoryToDiskImage doing nothing since "
      "the disk memory area hasn't been allocated yet.\n");
    goto NormalExit;
  }

  if (PathNameForCompressedDiskImage[0] == 0)
  {
    dprintf ("AGMSRAMDiskDevice: CopyMemoryToDiskImage doing nothing since "
      "there isn't a name for the compressed disk image file.\n");
    goto NormalExit;
  }

  FileBuffer = malloc (FILEBUFFERSIZE);
  if (FileBuffer == NULL)
  {
    dprintf ("AGMSRAMDiskDevice: CopyMemoryToDiskImage failed to allocate "
      "memory for the file writing buffer.\n");
    goto NormalExit;
  }

  if (UseSafeTemporaryFileWhenSavingCompressedDiskImage)
    OpenFileName = PathNameForTempDiskImage;
  else
    OpenFileName = PathNameForCompressedDiskImage;

  FileHandle = open (OpenFileName,
    O_CREAT | O_TRUNC | O_WRONLY | O_BINARY, 0666 /* file mode/permission */);
  if (FileHandle < 0)
  {
    ErrorCode = errno;
    dprintf ("AGMSRAMDiskDevice: CopyMemoryToDiskImage unable to create file "
      "\"%s\", error code %d (%s).\n",
      OpenFileName, ErrorCode, strerror (ErrorCode));
    goto NormalExit;
  }

  memset (&ZLibStream, 0, sizeof (ZLibStream));
  ZLibStream.zalloc = MyAllocForZLib;
  ZLibStream.zfree = MyFreeForZLib;

  /* The memory area acts as our input. */

  ZLibStream.next_in = MemoryAreaPntr;
  ZLibStream.avail_in = MemorySize;

  /* And we output to the file buffer. */

  ZLibStream.next_out = FileBuffer;
  ZLibStream.avail_out = FILEBUFFERSIZE;

  ErrorCode = deflateInit (&ZLibStream, CompressionLevel);
  if (ErrorCode != Z_OK)
  {
    dprintf ("AGMSRAMDiskDevice: CopyMemoryToDiskImage "
      "deflateInit for \"%s\" failed, code %d (%s).\n",
      OpenFileName, ErrorCode, ZLibStream.msg ? ZLibStream.msg : "?");
    goto NormalExit;
  }
  DeflateInitWasDone = true;

  /* Compress data until there is no more input. */

  while (true)
  {
    /* If the output buffer is full, write it to disk. */

    if (ZLibStream.avail_out == 0)
    {
      AmountWritten = write (FileHandle, FileBuffer, FILEBUFFERSIZE);
      if (AmountWritten < 0 || AmountWritten != FILEBUFFERSIZE)
      {
        ErrorCode = errno;
        dprintf ("AGMSRAMDiskDevice: CopyMemoryToDiskImage unable to write "
          "middle part of \"%s\", error code %d (%s).\n",
          OpenFileName, ErrorCode, strerror (ErrorCode));
        goto NormalExit;
      }

      ZLibStream.next_out = FileBuffer;
      ZLibStream.avail_out = FILEBUFFERSIZE;
    }

    /* Do a portion of the compression. */

    ErrorCode = deflate (&ZLibStream,
      (ZLibStream.avail_in == 0) ? Z_FINISH : Z_NO_FLUSH);

    /* If finished, write out the remaining partial output buffer and exit. */

    if (ErrorCode == Z_STREAM_END)
    {
      if (ZLibStream.avail_out > 0)
      {
        AmountWritten = write (FileHandle, FileBuffer,
          FILEBUFFERSIZE - ZLibStream.avail_out);
        if (AmountWritten < 0 ||
        AmountWritten != (FILEBUFFERSIZE - ZLibStream.avail_out))
        {
          ErrorCode = errno;
          dprintf ("AGMSRAMDiskDevice: CopyMemoryToDiskImage unable to write "
            "last part of \"%s\", error code %d (%s).\n",
            OpenFileName, ErrorCode, strerror (ErrorCode));
          goto NormalExit;
        }
      }
      break; /* Finished compressing all data. */
    }

    if (ErrorCode != Z_OK)
    {
      dprintf ("AGMSRAMDiskDevice: CopyMemoryToDiskImage "
        "deflate for \"%s\" failed after writing %ld bytes, code %d (%s).\n",
        OpenFileName, ZLibStream.total_out,
        ErrorCode, ZLibStream.msg ? ZLibStream.msg : "?");
      goto NormalExit;
    }
  }

  /* Now close the output file and rename it to be the final file if we
  were using a temporary file (simultaneously deletes the old file). */

  fsync (FileHandle);
  close (FileHandle);
  FileHandle = -1;

  if (UseSafeTemporaryFileWhenSavingCompressedDiskImage)
  {
    if (rename (PathNameForTempDiskImage, PathNameForCompressedDiskImage) < 0)
    {
      ErrorCode = errno;
      dprintf ("AGMSRAMDiskDevice: CopyMemoryToDiskImage "
        "renaming \"%s\" to \"%s\" failed, code %d (%s).\n",
        PathNameForTempDiskImage, PathNameForCompressedDiskImage,
        ErrorCode, strerror (ErrorCode));
      goto NormalExit;
    }
  }

  /* Successful. */

  ReturnCode = true;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("AGMSRAMDiskDevice: CopyMemoryToDiskImage "
    "successfully compressed %ld bytes to %ld bytes (%d%%) in file \"%s\".\n",
    ZLibStream.total_in, ZLibStream.total_out,
    (ZLibStream.total_in > 0)
    ? (int) (ZLibStream.total_out * 100LL /* LL for 64 bit ints */ /
      ZLibStream.total_in)
    : 0,
    PathNameForCompressedDiskImage);
#endif

  /* Clean up outstanding allocations. */

NormalExit:
  if (DeflateInitWasDone)
  {
    ErrorCode = deflateEnd (&ZLibStream);
    if (ErrorCode != Z_OK)
    {
      dprintf ("AGMSRAMDiskDevice: CopyMemoryToDiskImage deflateEnd for "
      "\"%s\" returned error code %d (%s).\n",
      OpenFileName, ErrorCode,
      ZLibStream.msg ? ZLibStream.msg : "?");
    }
  }
  if (FileHandle >= 0)
    close (FileHandle);
  if (FileBuffer != NULL)
    free (FileBuffer);
  return ReturnCode;
}



/******************************************************************************
 * Initial boot time initialisation.  For the RAM disk, we don't do much here,
 * other than checking for safe mode.  We also delete the previous RAM disk
 * memory area here, forcing a fresh start if the driver gets reloaded (which
 * happens only when developing the driver).
 */

status_t init_hardware (void)
{
  bool    SafeMode;
  void   *SettingsPntr;

  dprintf ("AGMSRAMDiskDevice: init_hardware compiled on " __DATE__ " at " __TIME__ "\n");
  dprintf ("AGMSRAMDiskDevice: init_hardware $Header: /boot/home/agmsmith/Programming/AGMSRAMDiskDevice/RCS/AGMSRAMDiskDevice.c,v 1.19 2001/05/14 21:50:44 agmsmith Exp $\n");

  /* Check to see if the operating system is running in safe mode, if it
  is, then don't initialise or do anything else. */

  SafeMode = false;
  SettingsPntr = load_driver_settings (B_SAFEMODE_DRIVER_SETTINGS);
  if (SettingsPntr != NULL)
  {
    SafeMode = get_driver_boolean_parameter (SettingsPntr,
      B_SAFEMODE_SAFE_MODE, false, false);
    unload_driver_settings (SettingsPntr);
  }

  if (SafeMode)
  {
    dprintf ("AGMSRAMDiskDevice: init_hardware returning failure, "
      "don't want to operate while in safe mode.\n");
    return B_ERROR;
  }

  /* Remove any existing RAM disk memory area.  This is mostly used for
  debugging since init_hardware only gets called once per boot, unless you
  change the date on the executable and close all files, then it will do
  an init_hardware again when you next open a file handle on this device. */

  FreeDiskContentsArea ();

  return B_OK;
}



/******************************************************************************
 * The driver has been freshly loaded into memory (or reloaded - all global
 * variables lost between driver uninit and the next driver init - which
 * happens when the device has zero open file handles).  Initialise things
 * used by the driver - like semaphores, and find the memory used for storing
 * the data.  Returning an error will force the immediate unloading of the
 * driver, code and all, so make sure nothing is allocated when an error is
 * returned.
 */

status_t init_driver (void)
{
  status_t  ErrorCode;

  ReadConfigurationParameters ();

  /* Find or allocate the memory area and get its address and size.
  Don't care if it failed to allocate memory, can reallocate later
  with B_SET_DEVICE_SIZE (meanwhile the user sees B_DEV_NOT_READY errors). */

  ErrorCode = AllocateDiskContentsArea ();

  /* Set up the general purpose semaphore. */

  MainSemaphoreID =
    create_sem (1 /* Allow 1 reader/writer */, MainSemaphoreName);
  if (MainSemaphoreID < 0)
  {
    ErrorCode = MainSemaphoreID;
    dprintf ("AGMSRAMDiskDevice: init_driver failed to create semaphore, "
      "error $%lX.\n", ErrorCode);
    return ErrorCode;
  }

  if ((ErrorCode = set_sem_owner (MainSemaphoreID, B_SYSTEM_TEAM)) < 0)
  {
    dprintf ("AGMSRAMDiskDevice: init_driver failed to change semaphore "
      "ownership, error $%lX.\n", ErrorCode);
    delete_sem (MainSemaphoreID);
    MainSemaphoreID = -1;
    return ErrorCode;
  }

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("AGMSRAMDiskDevice: init_driver created semaphore %d "
    "named \"%s\".\n", (int) MainSemaphoreID, MainSemaphoreName);
#endif

  /* Update a few housekeeping flags, but don't load in the new disk image
  at this point (wait until first "open") since loading the image takes a
  long time (decompressing a big file) and init_driver gets called
  several times during bootup just to find the names of the devices it
  exports, then uninit_driver gets called. */

  if (DiskIsRemoveable)
    DiskIsInserted = InsertDiskAtStartup;
  else
    DiskIsInserted = true;

  MemoryIsDirty = false;
  MediaLoadedSerialNo++;
  NumberOfOpenFiles = 0;
  RealFlushCountdown = FlushAtNumberOfOpenFilesHack;

  return B_OK;
}



/******************************************************************************
 * The driver is being unloaded, deallocate stuff we had allocated earlier
 * when the driver was loaded.  But don't deallocate the RAM memory area.
 * Note that this function doesn't get called when the system is going down
 * for a reboot, because the file system doesn't close its file handle
 * (or rather, each partition on the RAM disk leaves an open file handle).
 * However, the file system does do a flush near the end and since it
 * is the only handle open at that time, we dump the disk during that flush
 * too, as well as here where it should be done if things worked right.
 * It does work right if you reboot with the disk unmounted (such as after
 * using DriveSetup to format it, or if you have killed "Tracker").
 */

void uninit_driver (void)
{
  if (MemoryIsDirty)
  {
    if (CopyMemoryToDiskImage ())
      MemoryIsDirty = false;
  }

  if (MainSemaphoreID >= 0)
  {
    if (delete_sem (MainSemaphoreID) >= 0)
    {
#if DPRINT_NON_ERROR_MESSAGES
      dprintf ("AGMSRAMDiskDevice: uninit_driver deallocated semaphore %d.\n",
      (int) MainSemaphoreID);
#endif
    }
    else
    {
      dprintf ("AGMSRAMDiskDevice: uninit_driver failed to deallocate "
        "semaphore %d.\n", (int) MainSemaphoreID);
    }
    MainSemaphoreID = -1;
  }

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("AGMSRAMDiskDevice: uninit_driver finished.\n");
#endif
}



/******************************************************************************
 * Tell the driver system which devices we can handle.  The returned names
 * all get added to the /dev hierarchy and an association with this code is
 * made by the devfs (so the same code can run several devices).
 */

const char** publish_devices()
{
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("AGMSRAMDiskDevice: publish_devices returning \"%s\".\n",
    PublishedDeviceNames[0]);
#endif

  return PublishedDeviceNames;
}



/******************************************************************************
 * The user wants to open our device.  Make sure that the name matches our
 * device name then return a pointer to a newly allocated hunk of cookie data
 * to identify this particular opening.  The flags can be ORed bit combinations
 * of:
 *
 *  O_RDONLY        open for reading only
 *  O_WRONLY        open for writing only
 *  O_RDWR          open for reading and writing
 *  O_NONBLOCK      do not block on open or for data to become available
 *  O_APPEND        append on each write
 *  O_CREAT         create file if it does not exist
 *  O_TRUNC         truncate size to 0
 *  O_EXCL          error if create and file exists
 *  O_SYNC          perform syncronous I/O operations
 *  O_SHLOCK        atomically obtain a shared lock
 *  O_EXLOCK        atomically obtain an exclusive lock
 *  O_NOFOLLOW      if last path element is a symlink, don't follow it
 *  O_NOTRAVERSE    do not traverse leaf link
 *  O_CLOEXEC       close fd on exec
 *  O_TEXT          CR-LF translation
 *  O_BINARY        no translation
 *
 * We support O_RDONLY, O_WRONLY, O_RDWR.  We ignore the rest.
 */

static status_t
AGMSRAMDiskOpen (const char *DeviceName, uint32 Flags, void **CookiePntrPntr)
{
  status_t       ErrorCode;
  uint32         MaskedFlags;
  int            SavedNumberOfOpenFiles;
  CookieRecord   TempCookie;

  if (CookiePntrPntr == NULL)
  {
    dprintf ("AGMSRAMDiskDevice: attempt to open with NULL cookie.\n");
    return B_BAD_ADDRESS;
  }

  *CookiePntrPntr = NULL; /* In case of errors, default return is NULL. */

  if (DeviceName == NULL)
  {
    dprintf ("AGMSRAMDiskDevice: open - NULL device name specified.\n");
    return B_BAD_ADDRESS;
  }

  if (DiskIsReadOnly &&
  (((Flags & O_RWMASK) == O_RDWR) || ((Flags & O_RWMASK) == O_WRONLY)))
  {
    dprintf ("AGMSRAMDiskDevice: open - failed in attempt to open a "
      "read only disk for writing.\n");
    return B_PERMISSION_DENIED;
  }

  if (strcmp (DeviceName, PublishedDeviceNames[0]) != 0)
  {
    dprintf ("AGMSRAMDiskDevice: attempt to open with name \"%s\", which "
      "doesn't match our name of \"/dev/%s\".\n", DeviceName,
      PublishedDeviceNames[0]);
    return B_NAME_NOT_FOUND;
  }

  /* Critical section here, updating the serial number and loading the disk
  image if this is the first open. */

  if ((ErrorCode = acquire_sem (MainSemaphoreID)) < 0)
  {
    dprintf ("AGMSRAMDiskDevice: open unable to acquire semaphore, "
      "error code $%X.\n", (unsigned int) ErrorCode);
    return ErrorCode;
  }

  /* For the first "open", load the disk image into memory.  Not done in
  init_driver since that gets called several times during bootup, and would
  thus waste a lot of CPU time decompressing the disk image. */

  if (CookieCounter == 0 && DiskIsInserted)
  {
    CopyDiskImageToMemory ();
    MemoryIsDirty = false;
    MediaLoadedSerialNo++;
  }

  TempCookie.cookieCounter = ++CookieCounter;
  TempCookie.mediaSerial = MediaLoadedSerialNo;
  SavedNumberOfOpenFiles = ++NumberOfOpenFiles;

  /* The shutdown detection works by seeing that there are only N open files
  remaining (N is number of _mounted_ disk partitions), then it means that
  the Tracker has closed its file handle for monitoring the media status.
  At this point, we dump the entire RAM to a disk image on the Nth flush
  since each file system flushes before it dies (well, it may write a bit
  after it flushes, much to my annoyance, but if it is BFS then it will fix
  itself from the file system log when it remounts later on).  It may flush
  a few extra times too, but we'll do it only on the Nth (the later flushes
  don't always work anyways because the file system it is flushing to may
  not be working any more).  Also, while booting up, there may be a flush
  before the normal N+1 files are open, causing us to waste time, so waiting
  for the Nth flush helps there too. */

  if (NumberOfOpenFiles > FlushAtNumberOfOpenFilesHack)
    RealFlushCountdown = FlushAtNumberOfOpenFilesHack; /* Not shutting down. */

  release_sem (MainSemaphoreID);

  /* Extract the read / write open mode. */

  MaskedFlags = (Flags & O_RWMASK);
  TempCookie.writeAllowed = (MaskedFlags == O_WRONLY || MaskedFlags == O_RDWR);
  TempCookie.readAllowed = (MaskedFlags == O_RDONLY || MaskedFlags == O_RDWR);

  /* Set up other flags. */

  TempCookie.isOpen = true;

  /* Copy the cookie to a newly allocated piece of kernel memory, which is
  what the caller will hold on to for us while the file is open, and we
  will use to keep track of the status of this particular open file. */

  if ((*CookiePntrPntr = malloc (sizeof (TempCookie))) == NULL)
  {
    dprintf ("AGMSRAMDiskDevice: open unable to allocate memory for "
      "the cookie.\n");
    return B_DEV_NO_MEMORY;
  }

  /* Add the magic value and self pointer validity check fields.  These let
  us avoid trying to use garbage pointers to file descriptors (like those from
  a closed file the user is still using by mistake).  Though DevFS will
  probably check for them itself and shield us from them. */

  TempCookie.magicValue = 'AGMS';
  TempCookie.selfPntr = *CookiePntrPntr;

  memcpy (*CookiePntrPntr, &TempCookie, sizeof (TempCookie));

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("AGMSRAMDiskDevice #%d: opened with name \"%s\" and flags $%X, "
    "using R:%d W:%d%s, returning cookie $%X, %d files now open.\n",
    TempCookie.cookieCounter, DeviceName, (unsigned int) Flags,
    TempCookie.readAllowed, TempCookie.writeAllowed,
    (Flags & ~O_RWMASK) ? " (other flags ignored)" : "",
    (unsigned int) *CookiePntrPntr, SavedNumberOfOpenFiles);
#endif

  return B_OK;
}



/******************************************************************************
 * Read some data into the user's buffer.  Check for no media, and if the
 * disk is inserted, just do some range checking and copy the memory.
 */

static status_t AGMSRAMDiskRead (
void *Cookie, off_t Position, void *Buffer, size_t* AmountReadPntr)
{
  size_t        AmountToRead;
  CookiePointer CookiePntr = Cookie;
  status_t      ErrorCode;

  if (CookiePntr == NULL || Buffer == NULL || AmountReadPntr == NULL)
  {
    dprintf ("AGMSRAMDiskDevice #%d: read called with NULL pointers.\n",
      CookiePntr ? CookiePntr->cookieCounter : 0);
    return B_BAD_ADDRESS;
  }

  if (!CookiePntr->isOpen || CookiePntr->magicValue != 'AGMS' ||
  CookiePntr->selfPntr != CookiePntr)
  {
    dprintf ("AGMSRAMDiskDevice #%d: read called on a closed file handle.\n",
      CookiePntr->cookieCounter);
    *AmountReadPntr = 0;
    return B_FILE_ERROR;
  }

#if DPRINT_READ_AND_WRITE_OPERATIONS
  dprintf ("AGMSRAMDiskDevice #%d: read at position $%lX%08lX, amount %ld, "
    "buffer $%lX.\n", CookiePntr->cookieCounter,
    (unsigned long) (Position >> 32),
    (unsigned long) Position,
    *AmountReadPntr, (unsigned long) Buffer);
#endif

  if (!CookiePntr->readAllowed)
  {
    dprintf ("AGMSRAMDiskDevice #%d: read - file not originally opened with "
      "read access.\n", CookiePntr->cookieCounter);
    *AmountReadPntr = 0;
    return B_PERMISSION_DENIED;
  }

  if (Position < 0 || Position >= MemorySize)
  {
    dprintf ("AGMSRAMDiskDevice #%d: read past start or end of "
      "file attempted.\n", CookiePntr->cookieCounter);
    *AmountReadPntr = 0;
    return B_DEV_SEEK_ERROR;
  }

  AmountToRead = *AmountReadPntr;
  if (Position + AmountToRead > MemorySize)
    AmountToRead = MemorySize - Position;

  /* Critical section here, only allow one simultaneous reader or writer. */

  if ((ErrorCode = acquire_sem (MainSemaphoreID)) < 0)
  {
    dprintf ("AGMSRAMDiskDevice #%d: read - unable to acquire semaphore, "
      "error code $%X.\n", CookiePntr->cookieCounter,
      (unsigned int) ErrorCode);
    *AmountReadPntr = 0;
    return ErrorCode;
  }

  if (CookiePntr->mediaSerial != MediaLoadedSerialNo)
  {
    dprintf ("AGMSRAMDiskDevice #%d: read - failed because "
      "media was changed.\n", CookiePntr->cookieCounter);
    AmountToRead = 0;
    ErrorCode = B_DEV_MEDIA_CHANGED;
  }
  else if (!DiskIsInserted)
  {
    dprintf ("AGMSRAMDiskDevice #%d: read - failed because "
      "media isn't present.\n", CookiePntr->cookieCounter);
    AmountToRead = 0;
    ErrorCode = B_DEV_NO_MEDIA;
  }
  else if (MemoryAreaPntr == NULL)
  {
    dprintf ("AGMSRAMDiskDevice #%d: read - failed because "
      "storage hasn't been allocated (use B_SET_DEVICE_SIZE).\n",
      CookiePntr->cookieCounter);
    AmountToRead = 0;
    ErrorCode = B_DEV_NOT_READY;
  }
  else
  {
    memcpy (Buffer, MemoryAreaPntr + Position, AmountToRead);
    ErrorCode = B_OK;
  }

  release_sem (MainSemaphoreID);

  *AmountReadPntr = AmountToRead;
  return ErrorCode;
}



/******************************************************************************
 * Write some data from the user's buffer.  Check for the media being inserted
 * and then do some range checking and just copy the memory.
 */

static status_t AGMSRAMDiskWrite
(void* Cookie, off_t Position, const void* Buffer, size_t* AmountWrittenPntr)
{
  size_t        AmountToWrite;
  CookiePointer CookiePntr = Cookie;
  status_t      ErrorCode;

  if (CookiePntr == NULL || Buffer == NULL || AmountWrittenPntr == NULL)
  {
    dprintf ("AGMSRAMDiskDevice #%d: write called with NULL pointers.\n",
      CookiePntr ? CookiePntr->cookieCounter : 0);
    return B_BAD_ADDRESS;
  }

  if (!CookiePntr->isOpen || CookiePntr->magicValue != 'AGMS' ||
  CookiePntr->selfPntr != CookiePntr)
  {
    dprintf ("AGMSRAMDiskDevice #%d: write called on a closed file handle.\n",
      CookiePntr->cookieCounter);
    *AmountWrittenPntr = 0;
    return B_FILE_ERROR;
  }

#if DPRINT_READ_AND_WRITE_OPERATIONS
  dprintf ("AGMSRAMDiskDevice #%d: write at position $%lX%08lX, amount %ld, "
    "buffer $%lX.\n", CookiePntr->cookieCounter,
    (unsigned long) (Position >> 32),
    (unsigned long) Position,
    *AmountWrittenPntr, (unsigned long) Buffer);
#endif

  if (!CookiePntr->writeAllowed)
  {
    dprintf ("AGMSRAMDiskDevice #%d: write - file not originally opened "
      "with write access.\n", CookiePntr->cookieCounter);
    *AmountWrittenPntr = 0;
    return B_PERMISSION_DENIED;
  }

  if (Position < 0 || Position >= MemorySize)
  {
    dprintf ("AGMSRAMDiskDevice #%d: write past start or end of "
      "memory attempted.\n", CookiePntr->cookieCounter);
    *AmountWrittenPntr = 0;
    return B_DEV_SEEK_ERROR;
  }

  AmountToWrite = *AmountWrittenPntr;
  if (Position + AmountToWrite > MemorySize)
    AmountToWrite = MemorySize - Position;

  /* Critical section here, only allow one simultaneous reader or writer. */

  if ((ErrorCode = acquire_sem (MainSemaphoreID)) < 0)
  {
    dprintf ("AGMSRAMDiskDevice #%d: write - unable to acquire semaphore, "
      "error code $%X.\n", CookiePntr->cookieCounter,
      (unsigned int) ErrorCode);
    *AmountWrittenPntr = 0;
    return ErrorCode;
  }

  if (CookiePntr->mediaSerial != MediaLoadedSerialNo)
  {
    dprintf ("AGMSRAMDiskDevice #%d: write - failed because "
      "media was changed.\n", CookiePntr->cookieCounter);
    AmountToWrite = 0;
    ErrorCode = B_DEV_MEDIA_CHANGED;
  }
  else if (!DiskIsInserted)
  {
    dprintf ("AGMSRAMDiskDevice #%d: write - failed because "
      "media isn't present.\n", CookiePntr->cookieCounter);
    AmountToWrite = 0;
    ErrorCode = B_DEV_NO_MEDIA;
  }
  else if (MemoryAreaPntr == NULL)
  {
    dprintf ("AGMSRAMDiskDevice #%d: write - failed because "
      "storage hasn't been allocated (use B_SET_DEVICE_SIZE).\n",
      CookiePntr->cookieCounter);
    AmountToWrite = 0;
    ErrorCode = B_DEV_NOT_READY;
  }
  else if (DiskIsReadOnly)
  {
    dprintf ("AGMSRAMDiskDevice #%d: write - failed because disk "
      "is read only.\n", CookiePntr->cookieCounter);
    AmountToWrite = 0;
    ErrorCode = B_DEV_WRITE_ERROR;
  }
  else if (DiskIsWriteOnce &&
  !CheckForEmptyWriteOnceData (Position, AmountToWrite))
  {
    /* For write once disks, return an error if writing to a non-empty series
    of sectors, where empty means they are all zeroes.  Alternatively we could
    logical OR the bits to the existing bits, but an error makes write once
    file system bugs more immediately obvious. */

    dprintf ("AGMSRAMDiskDevice #%d: write - failed because write-once disk "
      "has data already in area to be written (position %d, length %d).\n",
      CookiePntr->cookieCounter, (int) Position, (int) AmountToWrite);
    AmountToWrite = 0;
    ErrorCode = B_DEV_WRITE_ERROR;
  }
  else
  {
    memcpy (MemoryAreaPntr + Position, Buffer, AmountToWrite);
    MemoryIsDirty = true;
    ErrorCode = B_OK;
  }

  release_sem (MainSemaphoreID);

  *AmountWrittenPntr = AmountToWrite;
  return ErrorCode;
}



/******************************************************************************
 * Respond to miscellaneous commands, such as queries for disk size.  Actually
 * this is the most complex part of the RAM disk.
 */

static status_t AGMSRAMDiskControl
(void* Cookie, uint32 OpCode, void* Buffer, size_t Length)
{
  CookiePointer CookiePntr = Cookie;
  status_t      ErrorCode = B_OK;

  if (Cookie == NULL)
  {
    dprintf ("AGMSRAMDiskDevice #0: control called with NULL file pointer.\n");
    return B_BAD_ADDRESS;
  }

  if (!CookiePntr->isOpen || CookiePntr->magicValue != 'AGMS' ||
  CookiePntr->selfPntr != CookiePntr)
  {
    dprintf ("AGMSRAMDiskDevice #%d: control called on a closed "
      "file handle.\n", CookiePntr->cookieCounter);
    return B_FILE_ERROR;
  }

  /* Since this is a critical section, all code paths should flow to the end
  of this function, rather than returning in the middle, otherwise the
  semaphore won't get released properly. */

  if ((ErrorCode = acquire_sem (MainSemaphoreID)) < 0)
  {
    dprintf ("AGMSRAMDiskDevice #%d: control - unable to acquire semaphore, "
      "error code $%X.\n", CookiePntr->cookieCounter,
      (unsigned int) ErrorCode);
    return ErrorCode;
  }

  ErrorCode = B_OK;

  switch (OpCode)
  {
    case B_GET_DEVICE_SIZE: /* Returns 32 bit size of device. */
      if (Buffer == NULL)
        goto NullBufferErrorExit;
      if (DiskIsInserted && MemoryAreaPntr != NULL)
        * (size_t *) Buffer = MemorySize;
      else /* Similar to B_GET_GEOMETRY, report zero if no media. */
        * (size_t *) Buffer = 0;
#if DPRINT_NON_ERROR_MESSAGES
      dprintf ("AGMSRAMDiskDevice #%d: control - B_GET_DEVICE_SIZE - "
        "returning %ld.\n", CookiePntr->cookieCounter, * (size_t *) Buffer);
#endif
      break;


    case B_SET_DEVICE_SIZE: /* Resize the RAM disk, if inserted. */
      if (Buffer == NULL)
        goto NullBufferErrorExit;
      if (!DiskIsInserted)
      {
        dprintf ("AGMSRAMDiskDevice #%d: control - B_SET_DEVICE_SIZE "
          "failed because media isn't present.\n", CookiePntr->cookieCounter);
        ErrorCode = B_DEV_NO_MEDIA;
      }
      else /* Do the resize. */
      {
        InitialMemoryByteSize = * (size_t *) Buffer;
        if (InitialMemoryByteSize < FAKE_BLOCK_SIZE)
          InitialMemoryByteSize = FAKE_BLOCK_SIZE;
        InitialMemoryByteSize = InitialMemoryByteSize /
          FAKE_BLOCK_SIZE * FAKE_BLOCK_SIZE;
#if DPRINT_NON_ERROR_MESSAGES
        dprintf ("AGMSRAMDiskDevice #%d: control - B_SET_DEVICE_SIZE, "
          "resizing to %ld bytes (caller requested %ld).\n",
          CookiePntr->cookieCounter, InitialMemoryByteSize,
          * (size_t *) Buffer);
#endif
        FreeDiskContentsArea ();
        ErrorCode = AllocateDiskContentsArea ();
        if (ErrorCode < 0)
        {
          dprintf ("AGMSRAMDiskDevice #%d: control - B_SET_DEVICE_SIZE failed, "
            "unable to allocate memory, old contents lost, "
            "you should try again.\n",
            CookiePntr->cookieCounter);
          /* Note that we continue on with MemoryAreaPntr == NULL, this will
          cause B_DEV_NOT_READY errors when anyone tries to read / write. */
        }
        MemoryIsDirty = true;
        MediaLoadedSerialNo++; /* Media is changed for everybody else. */
        CookiePntr->mediaSerial = MediaLoadedSerialNo;
      }
      break;


    case B_SET_BLOCKING_IO: /* We always run in blocking mode, so OK. */
#if DPRINT_NON_ERROR_MESSAGES
      dprintf ("AGMSRAMDiskDevice #%d: control - B_SET_BLOCKING_IO OK.\n",
        CookiePntr->cookieCounter);
#endif
      break;


    case B_GET_READ_STATUS: /* Returns TRUE if we can read without blocking. */
      if (Buffer == NULL)
        goto NullBufferErrorExit;
#if DPRINT_NON_ERROR_MESSAGES
      dprintf ("AGMSRAMDiskDevice #%d: control - B_GET_READ_STATUS - "
        "returning false (read might block).\n", CookiePntr->cookieCounter);
#endif
      * (bool *) Buffer = false;
      break;


    case B_GET_WRITE_STATUS: /* Returns TRUE if we can write without blocking. */
      if (Buffer == NULL)
        goto NullBufferErrorExit;
#if DPRINT_NON_ERROR_MESSAGES
      dprintf ("AGMSRAMDiskDevice #%d: control - B_GET_WRITE_STATUS - "
        "returning false (write might block).\n", CookiePntr->cookieCounter);
#endif
      * (bool *) Buffer = false;
      break;


    case B_GET_GEOMETRY: /* Describe our RAM drive. */
      if (Buffer == NULL)
        goto NullBufferErrorExit;
      {
        device_geometry *GeometryPntr = Buffer;

        if (DiskIsInserted && MemoryAreaPntr != NULL)
        {
          GeometryPntr->bytes_per_sector = FAKE_BLOCK_SIZE;
          GeometryPntr->sectors_per_track = 1;
          GeometryPntr->head_count = 1;
          GeometryPntr->cylinder_count =
            MemorySize / (FAKE_BLOCK_SIZE * GeometryPntr->sectors_per_track *
            GeometryPntr->head_count);
        }
        else /* No disk present, report zero sizes, even blocksize. */
        {
          GeometryPntr->bytes_per_sector = 0;
          GeometryPntr->sectors_per_track = 0;
          GeometryPntr->head_count = 0;
          GeometryPntr->cylinder_count = 0;
        }

        GeometryPntr->device_type = B_DISK;
        GeometryPntr->removable = DiskIsRemoveable;
        GeometryPntr->read_only = DiskIsReadOnly;
        GeometryPntr->write_once = DiskIsWriteOnce;

#if DPRINT_NON_ERROR_MESSAGES
        dprintf ("AGMSRAMDiskDevice #%d: control - B_GET_%sGEOMETRY is %ld "
          "bytes per sector, %ld sectors per track, %ld heads, %ld cylinders "
          "(total %lu), removeable %d, read only %d, write once %d.\n",
          CookiePntr->cookieCounter,
          (OpCode == B_GET_BIOS_GEOMETRY) ? "BIOS_" : "",
          GeometryPntr->bytes_per_sector,
          GeometryPntr->sectors_per_track,
          GeometryPntr->head_count,
          GeometryPntr->cylinder_count,
          GeometryPntr->bytes_per_sector *
          GeometryPntr->sectors_per_track *
          GeometryPntr->head_count * GeometryPntr->cylinder_count,
          (int) GeometryPntr->removable,
          (int) GeometryPntr->read_only,
          (int) GeometryPntr->write_once);
#endif
      }
      break;


    case B_FORMAT_DEVICE: /* Erase it completely with zeroes. */
      if (Buffer == NULL)
        goto NullBufferErrorExit;
      if (!DiskIsInserted)
      {
        dprintf ("AGMSRAMDiskDevice #%d: control - B_FORMAT_DEVICE "
          "failed because media isn't present.\n", CookiePntr->cookieCounter);
        ErrorCode = B_DEV_NO_MEDIA;
      }
      else if (MemoryAreaPntr == NULL)
      {
        dprintf ("AGMSRAMDiskDevice #%d: control - B_FORMAT_DEVICE "
          "failed because storage doesn't exist (use B_SET_DEVICE_SIZE).\n",
          CookiePntr->cookieCounter);
        ErrorCode = B_DEV_NOT_READY;
      }
      else if (DiskIsReadOnly)
      {
        dprintf ("AGMSRAMDiskDevice #%d: control - B_FORMAT_DEVICE "
          "failed because disk is read only.\n",
          CookiePntr->cookieCounter);
        ErrorCode = B_DEV_WRITE_ERROR;
      }
      else /* Do the format. */
      {
#if DPRINT_NON_ERROR_MESSAGES
        dprintf ("AGMSRAMDiskDevice #%d: control - B_FORMAT_DEVICE, "
          "full format:%d.\n", CookiePntr->cookieCounter,
          (int) (* (bool *) Buffer));
#endif
        if (* (bool *) Buffer)
          memset (MemoryAreaPntr, 0, MemorySize);
        else /* Only do a partial format. */
          memset (MemoryAreaPntr, 0, (MemorySize > 1024) ? 1024 : MemorySize);

        MemoryIsDirty = true;
        MediaLoadedSerialNo++; /* Media is changed for everybody else. */
        CookiePntr->mediaSerial = MediaLoadedSerialNo;
      }
      break;


    case B_EJECT_DEVICE:
      if (DiskIsRemoveable)
      {
        if (DiskIsInserted)
        {
          if (MemoryIsDirty)
          {
            if (CopyMemoryToDiskImage ())
              MemoryIsDirty = false;  /* Memory is same as stuff on disk. */
          }
          MediaLoadedSerialNo++;
          DiskIsInserted = false;
        }
#if DPRINT_NON_ERROR_MESSAGES
        dprintf ("AGMSRAMDiskDevice #%d: control - B_EJECT_DEVICE done.\n",
          CookiePntr->cookieCounter);
#endif
      }
      else
      {
        ErrorCode = B_DEV_INVALID_IOCTL;
        dprintf ("AGMSRAMDiskDevice #%d: control - B_EJECT_DEVICE invalid for "
          "non-removeable devices.\n", CookiePntr->cookieCounter);
      }
      break;


    case B_GET_ICON:
      if (Buffer == NULL)
        goto NullBufferErrorExit;
      {
        device_icon *IconInfo = Buffer;
        int x, y;

#if DPRINT_NON_ERROR_MESSAGES
        dprintf ("AGMSRAMDiskDevice #%d: control - B_GET_ICON of size %d, "
          "buffer $%lX.\n", CookiePntr->cookieCounter,
           (int) IconInfo->icon_size, (unsigned long) IconInfo->icon_data);
#endif
        if (IconInfo->icon_size == 16)
          memcpy (IconInfo->icon_data, DiskIcon16, 16 * 16);
        else if (IconInfo->icon_size == 32)
          memcpy (IconInfo->icon_data, DiskIcon32, 32 * 32);
        else /* Make an algorithmic icon. */
        {
          for (x = 0; x < IconInfo->icon_size; x++)
            for (y = 0; y < IconInfo->icon_size; y++)
            {
              ((unsigned char *) IconInfo->icon_data)
                [x + y * IconInfo->icon_size] =
                x + y + x * y;
            }
        }
      }
      break;


    case B_GET_MEDIA_STATUS:
      if (Buffer == NULL)
        goto NullBufferErrorExit;
      {
        status_t *ReturnValue = Buffer;

        if (CookiePntr->mediaSerial != MediaLoadedSerialNo)
        {
#if DPRINT_NON_ERROR_MESSAGES
          dprintf ("AGMSRAMDiskDevice #%d: control - B_GET_MEDIA_STATUS "
            "returns B_DEV_MEDIA_CHANGED & clears change flag.\n",
            CookiePntr->cookieCounter);
#endif
          CookiePntr->mediaSerial = MediaLoadedSerialNo;
          *ReturnValue = B_DEV_MEDIA_CHANGED;
        }
        else if (!DiskIsInserted)
        {
#if DPRINT_NON_ERROR_MESSAGES
          dprintf ("AGMSRAMDiskDevice #%d: control - B_GET_MEDIA_STATUS "
            "returns B_DEV_DOOR_OPEN.\n", CookiePntr->cookieCounter);
#endif
          *ReturnValue = B_DEV_DOOR_OPEN;
        }
        else if (MemoryAreaPntr == NULL)
        {
#if DPRINT_NON_ERROR_MESSAGES
          dprintf ("AGMSRAMDiskDevice #%d: control - B_GET_MEDIA_STATUS "
            "returns B_DEV_NOT_READY since storage hasn't been allocated.\n",
            CookiePntr->cookieCounter);
#endif
          *ReturnValue = B_DEV_NOT_READY;
        }
        else
        {
#if DPRINT_NON_ERROR_MESSAGES
          dprintf ("AGMSRAMDiskDevice #%d: control - B_GET_MEDIA_STATUS "
            "returns B_NO_ERROR (drive ready).\n", CookiePntr->cookieCounter);
#endif
          *ReturnValue = B_NO_ERROR;
        }
      }
      break;


    case B_LOAD_MEDIA:
      if (DiskIsRemoveable)
      {
        if (!DiskIsInserted)
        {
          CopyDiskImageToMemory ();
          MemoryIsDirty = false; /* Stuff in memory same as stuff on disk. */
          MediaLoadedSerialNo++;
          DiskIsInserted = true;
        }
#if DPRINT_NON_ERROR_MESSAGES
        dprintf ("AGMSRAMDiskDevice #%d: control - B_LOAD_MEDIA done.\n",
          CookiePntr->cookieCounter);
#endif
      }
      else
      {
        ErrorCode = B_DEV_INVALID_IOCTL;
        dprintf ("AGMSRAMDiskDevice #%d: control - B_LOAD_MEDIA invalid for "
          "non-removeable devices.\n", CookiePntr->cookieCounter);
      }
      break;


    case B_SET_UNINTERRUPTABLE_IO:
#if DPRINT_NON_ERROR_MESSAGES
      dprintf ("AGMSRAMDiskDevice #%d: control - B_SET_UNINTERRUPTABLE_IO "
        "done.\n", CookiePntr->cookieCounter);
#endif
      break;


    case B_FLUSH_DRIVE_CACHE:
      if (!DiskIsInserted)
      {
        dprintf ("AGMSRAMDiskDevice #%d: control - B_FLUSH_DRIVE_CACHE "
          "failed because media isn't present.\n", CookiePntr->cookieCounter);
        ErrorCode = B_DEV_NO_MEDIA;
      }
      else
      {
        /* Because the system doesn't shut down properly, leaving one file
        always open (the file system accessing the device, guess it doesn't
        unmount it for BFS, same for DOS), write out the data when it flushes
        when there is only one remaining file open.  Well, actually there is
        one open file per partition, so let the user specify the number of
        open files needed to trigger the flush.  And only flush on the Nth
        flush when shutdown is active, so we don't repeatedly waste time
        compressing the whole disk image for each partition on the disk. */

        if (NumberOfOpenFiles == FlushAtNumberOfOpenFilesHack)
        {
          /* Probably shutting down, only N files open, not the usual N+1,
          though it could be starting up too. */

          if (--RealFlushCountdown == 0 && MemoryIsDirty)
          {
            /* This is the Nth flush during shutdown, dump the data. */

            if (CopyMemoryToDiskImage ())
              MemoryIsDirty = false;
          }
        }
#if DPRINT_NON_ERROR_MESSAGES
        dprintf ("AGMSRAMDiskDevice #%d: control - B_FLUSH_DRIVE_CACHE done "
          "(%d files open, flush countdown now %d).\n",
          CookiePntr->cookieCounter, NumberOfOpenFiles, RealFlushCountdown);
#endif
      }
      break;


    default:
      ErrorCode = B_DEV_INVALID_IOCTL;
      dprintf ("AGMSRAMDiskDevice #%d: control - opcode %lu is "
        "unimplemented.\n", CookiePntr->cookieCounter, OpCode);
      break;
  }
  goto NormalExit;

NullBufferErrorExit:
  dprintf ("AGMSRAMDiskDevice #%d: control operation #%lu failed, "
    "it needs a buffer and the caller provided a NULL pointer.\n",
    CookiePntr->cookieCounter, OpCode);
  ErrorCode = B_BAD_ADDRESS;
  goto NormalExit;

NormalExit:
  release_sem (MainSemaphoreID);
  return ErrorCode;
}



/******************************************************************************
 * Close an instance of the device.  Unblock any other pending reads (usually
 * done by deleting the semaphore).  Not much to do for us.
 */

static status_t AGMSRAMDiskClose (void* Cookie)
{
  CookiePointer CookiePntr = Cookie;
  status_t      ErrorCode;

  if (Cookie == NULL)
  {
    dprintf ("AGMSRAMDiskDevice #0: close called with NULL file pointer.\n");
    return B_BAD_ADDRESS;
  }

  if (!CookiePntr->isOpen || CookiePntr->magicValue != 'AGMS' ||
  CookiePntr->selfPntr != CookiePntr)
  {
    dprintf ("AGMSRAMDiskDevice #%d: close called on a closed "
      "file handle.\n", CookiePntr->cookieCounter);
    return B_FILE_ERROR;
  }

  /* Note that using the semaphores here is pointless since we
  don't do anything with the shared variables.  But maybe in the future... */

  if ((ErrorCode = acquire_sem (MainSemaphoreID)) < 0)
  {
    dprintf ("AGMSRAMDiskDevice #%d: close - unable to acquire semaphore, "
      "error code $%X.\n", CookiePntr->cookieCounter,
      (unsigned int) ErrorCode);
    return ErrorCode;
  }

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("AGMSRAMDiskDevice #%d: closing file.\n",
    CookiePntr->cookieCounter);
#endif

  CookiePntr->isOpen = false; /* Stop future operations on this file. */

  release_sem (MainSemaphoreID);
  return B_OK;
}



/******************************************************************************
 * Free data structures allocated during opening of a particular instance of
 * this driver.  Not much to do, except for the last open file (dump memory
 * to the disk image file, uninit doesn't always work so we do it here).
 */

static status_t AGMSRAMDiskFree (void* Cookie)
{
  CookiePointer CookiePntr = Cookie;
  int           ErrorCode;
  bool          GotSemaphore;

  if (Cookie == NULL)
  {
    dprintf ("AGMSRAMDiskDevice: free called with NULL file pointer.\n");
    return B_BAD_ADDRESS;
  }

  if (CookiePntr->magicValue != 'AGMS' ||
  CookiePntr->selfPntr != CookiePntr)
  {
    dprintf ("AGMSRAMDiskDevice #%d: free called on a dead "
      "file handle.\n", CookiePntr->cookieCounter);
    return B_FILE_ERROR;
  }

  ErrorCode = acquire_sem (MainSemaphoreID);
  GotSemaphore = (ErrorCode >= 0);
  if (!GotSemaphore)
  {
    dprintf ("AGMSRAMDiskDevice #%d: free - unable to acquire semaphore, "
      "error code $%X, but continuing to free the file in an unsafe manner.\n",
      CookiePntr->cookieCounter, (unsigned int) ErrorCode);
  }

  NumberOfOpenFiles--;
  CookiePntr->magicValue = 'Dead'; /* So it can't be reused by accident. */
  CookiePntr->selfPntr = NULL;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("AGMSRAMDiskDevice #%d: freeing file, now have %d open files.\n",
    CookiePntr->cookieCounter, NumberOfOpenFiles);
#endif

  free (CookiePntr);

  if (GotSemaphore)
    release_sem (MainSemaphoreID);

  return B_OK;
}



/******************************************************************************
 * A table of our hook (callback) functions so that the operating system and
 * users can call them.
 */

device_hooks AGMSRAMDiskHooks =
{
  AGMSRAMDiskOpen,
  AGMSRAMDiskClose,
  AGMSRAMDiskFree,
  AGMSRAMDiskControl,
  AGMSRAMDiskRead,
  AGMSRAMDiskWrite,
  NULL, /* We don't implement the select hooks. */
  NULL,
  NULL, /* Don't implement the scattered buffer read and write. */
  NULL
};



/******************************************************************************
 * For a given device, return pointers to the functions used to handle it.
 * Return NULL if it is asking about some other device.
 */

device_hooks* find_device (const char* DeviceName)
{
  if (DeviceName == NULL ||
  strcmp (DeviceName, PublishedDeviceNames[0]) != 0)
  {
    dprintf ("AGMSRAMDiskDevice: find_device with name \"%s\", which doesn't "
      "match our name of \"/dev/%s\".\n", DeviceName,
      PublishedDeviceNames[0]);
    return NULL;
  }

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("AGMSRAMDiskDevice: find_device called with name \"%s\", "
    "now returning our hook functions.\n", DeviceName);
#endif

  return &AGMSRAMDiskHooks;
}
